__all__ = [
    "__title__", "__summary__", "__uri__", "__author__", "__email__", "__version__",
]

__title__ = "covid-shared"
__summary__ = "Shared utilities for the covid model pipeline."
__uri__ = "https://github.com/ihmeuw/covid-shared"

__version__ = '2.0.0'

__author__ = "James Collins"
__email__ = "collijk@uw.edu"
